/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package it.iit.genomics.cru.isomirage;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Hashtable;

/**
 * @(#)IsoformFactory.java
 * 11 Sept 2014
 * Copyright 2014 Computational Research Unit of IIT@SEMM. All rights reserved.
 * Use is subject to MIT license terms.
 * 
 * Generates isoforms for miRNAs
 * 
 * @author Heiko Muller
 * @version 1.0
 * @since 1.0
 */
public class IsoformFactory {

    /**
     * Input file holding base miRNAs downloaded for example from mirBase
     */
    File input;
    
    /**
     * holds base miRNAs as String in format: name \t sequence
     */
    ArrayList<String> mirnas;
    
    /**
     * list of extensions added to 3' end and 5' end of base miRNA
     */
    String[] extensions = {
        "A",
        "C",
        "T",
        "G",
        "AA",
        "AC",
        "AG",
        "AT",
        "CA",
        "CC",
        "CG",
        "CT",
        "GA",
        "GC",
        "GG",
        "GT",
        "TA",
        "TC",
        "TG",
        "TT",
        "AAA",
        "AAC",
        "AAG",
        "AAT",
        "ACA",
        "ACC",
        "ACG",
        "ACT",
        "AGA",
        "AGC",
        "AGG",
        "AGT",
        "ATA",
        "ATC",
        "ATG",
        "ATT",
        "CAA",
        "CAC",
        "CAG",
        "CAT",
        "CCA",
        "CCC",
        "CCG",
        "CCT",
        "CGA",
        "CGC",
        "CGG",
        "CGT",
        "CTA",
        "CTC",
        "CTG",
        "CTT",
        "GAA",
        "GAC",
        "GAG",
        "GAT",
        "GCA",
        "GCC",
        "GCG",
        "GCT",
        "GGA",
        "GGC",
        "GGG",
        "GGT",
        "GTA",
        "GTC",
        "GTG",
        "GTT",
        "TAA",
        "TAC",
        "TAG",
        "TAT",
        "TCA",
        "TCC",
        "TCG",
        "TCT",
        "TGA",
        "TGC",
        "TGG",
        "TGT",
        "TTA",
        "TTC",
        "TTG",
        "TTT",};

    /**
     * Constructor of IsoformFactory
     * @param f base miRNAs in genome format
     */
    public IsoformFactory(File f) {
        input = f;
    }

    /**
     * Reads input genome file
     */
    private void readInput() {
        try {
            FileReader fr = new FileReader(input);
            BufferedReader br = new BufferedReader(fr);
            mirnas = new ArrayList<String>();
            String line = "";
            while ((line = br.readLine()) != null) {
                if(line.length() > 0){
                    mirnas.add(line);
                }
            }
        } catch (Exception e) {
        }
    }

    /**
     * Generates Isoforms. Isoforms are put into a hash table to ensure that no sequence is duplicated.
     */
    public void generateIsoforms() {
        readInput();
        Hashtable<String, String> ht = new Hashtable<String, String>();
        for (String s : mirnas) {
            String id = s.split("\t")[0].trim();
            String seq = s.split("\t")[1].trim();
            String[][] mir = formatInputIDs(id, seq);
            String[][] p3 = add3PrimeExtensions(id, seq);
            String[][] p5 = add5PrimeExtensions(id, seq);
            String[][] trim = trim(id, seq);
            if (!ht.containsKey(mir[0][1])) {
                ht.put(mir[0][1], mir[0][0]);
            }
            for (int i = 0; i < p3.length; i++) {
                if (!ht.containsKey(p3[i][1])) {
                    ht.put(p3[i][1], p3[i][0]);
                }
            }
            for (int i = 0; i < p5.length; i++) {
                if (!ht.containsKey(p5[i][1])) {
                    ht.put(p5[i][1], p5[i][0]);
                }
            }
            if (trim != null) {
                for (int i = 0; i < trim.length; i++) {
                    if (!ht.containsKey(trim[i][1])) {
                        ht.put(trim[i][1], trim[i][0]);
                    }
                }
            }
        }
        //dumpHash(ht);
        saveGenome(ht);
        saveFasta(ht);
    }

    /**
     * Formats input base miRNA identifier
     * @param id
     * @param seq
     * @return String[][]
     */
    private String[][] formatInputIDs(String id, String seq) {
        String[][] result = new String[1][2];
        result[0][1] = seq;
        result[0][0] = id + "_miRNA_" + result[0][1];
        return result;
    }

    /**
     * Adds extensions to 3'end of base miRNA
     * @param id
     * @param seq
     * @return String[][]
     */
    private String[][] add3PrimeExtensions(String id, String seq) {
        String[][] result = new String[84][2];
        for (int i = 0; i < extensions.length; i++) {
            result[i][1] = seq + extensions[i];
            result[i][0] = id + "_" + extensions[i] + "_3prime_" + result[i][1];
        }
        return result;
    }

    /**
     * Adds extensions to 5'end of base miRNA
     * @param id
     * @param seq
     * @return String[][]
     */
    private String[][] add5PrimeExtensions(String id, String seq) {
        String[][] result = new String[84][2];
        for (int i = 0; i < extensions.length; i++) {
            result[i][1] = extensions[i] + seq;
            result[i][0] = id + "_" + extensions[i] + "_5prime_" + result[i][1];
        }
        return result;
    }

    /**
     * Trims base miRNA one by one down to 18 nucleotides of length
     * @param id
     * @param seq
     * @return String[][]
     */
    private String[][] trim(String id, String seq) {
        if (seq.length() > 18) {
            String[][] result = new String[seq.length() - 18][2];
            for (int i = 1; i <= seq.length() - 18; i++) {
                result[i - 1][1] = seq.substring(0, seq.length() - i);
                result[i - 1][0] = id + "_trim" + i + "_" + result[i - 1][1];
            }
            return result;
        }
        return null;
    }

    /**
     * Dumps hash to console
     * @param ht 
     */
    private void dumpHash(Hashtable<String, String> ht) {
        Object[] keys = ht.keySet().toArray();
        for (int i = 0; i < keys.length; i++) {
            String id = ht.get(keys[i]);
            System.out.println(id + "\t" + (String) keys[i]);
        }
    }

    /**
     * Saves isoforms in genome format to file
     * @param ht 
     */
    private void saveGenome(Hashtable<String, String> ht) {
        Object[] keys = ht.keySet().toArray();
        try {
            FileWriter fwr = new FileWriter(input.getParent() + File.separator + input.getName() + ".genome");
            for (int i = 0; i < keys.length; i++) {
                String id = ht.get(keys[i]);
                fwr.write(id + "\t" + (String) keys[i] + "\r\n");
            }
            fwr.flush();
            fwr.close();
        } catch (Exception e) {
        }
    }

    /**
     * Saves isoforms in fasta format to file
     * @param ht 
     */
    private void saveFasta(Hashtable<String, String> ht) {
        Object[] keys = ht.keySet().toArray();
        try {
            FileWriter fwr = new FileWriter(input.getParent() + File.separator + input.getName() + ".fasta");
            for (int i = 0; i < keys.length; i++) {
                String id = ht.get(keys[i]);
                fwr.write(">" + id + "\r\n" + (String) keys[i] + "\r\n");
            }
            fwr.flush();
            fwr.close();
        } catch (Exception e) {
        }
    }

    /**
     * main method for testing
     * @param a 
     */
    /*
    public static void main(String[] a) {
        //IsoformFactory factory = new IsoformFactory(new File("C:\\Users\\hmuller\\Documents\\docs\\mirna\\hsa21\\hsa21.txt"));
        //factory.generateIsoforms();
    }
    */
}
